import sys

import pytest

from piccolo.engine import engine_finder
from piccolo.testing.test_case import (
    AsyncTableTest,
    AsyncTransactionTest,
    TableTest,
)
from tests.example_apps.music.tables import Band, Manager


class TestTableTest(TableTest):
    """
    Make sure the tables are created automatically.
    """

    tables = [Band, Manager]

    async def test_tables_created(self):
        self.assertTrue(Band.table_exists().run_sync())
        self.assertTrue(Manager.table_exists().run_sync())


class TestAsyncTableTest(AsyncTableTest):
    """
    Make sure the tables are created automatically in async tests.
    """

    tables = [Band, Manager]

    async def test_tables_created(self):
        self.assertTrue(await Band.table_exists())
        self.assertTrue(await Manager.table_exists())


@pytest.mark.skipif(sys.version_info <= (3, 11), reason="Python 3.11 required")
class TestAsyncTransaction(AsyncTransactionTest):
    """
    Make sure that the test exists within a transaction.
    """

    async def test_transaction_exists(self):
        db = engine_finder()
        assert db is not None
        self.assertTrue(db.transaction_exists())


@pytest.mark.skipif(sys.version_info <= (3, 11), reason="Python 3.11 required")
class TestAsyncTransactionRolledBack(AsyncTransactionTest):
    """
    Make sure that the changes get rolled back automatically.
    """

    async def asyncTearDown(self):
        await super().asyncTearDown()

        assert Manager.table_exists().run_sync() is False

    async def test_insert_data(self):
        await Manager.create_table()

        manager = Manager({Manager.name: "Guido"})
        await manager.save()



from __future__ import annotations

import typing as t

from piccolo.custom_types import Combinable, TableInstance
from piccolo.query.base import Query
from piccolo.query.mixins import (
    AddDelegate,
    OnConflictAction,
    OnConflictDelegate,
    ReturningDelegate,
)
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns.base import Column
    from piccolo.table import Table


class Insert(
    t.Generic[TableInstance], Query[TableInstance, t.List[t.Dict[str, t.Any]]]
):
    __slots__ = ("add_delegate", "on_conflict_delegate", "returning_delegate")

    def __init__(
        self, table: t.Type[TableInstance], *instances: TableInstance, **kwargs
    ):
        super().__init__(table, **kwargs)
        self.add_delegate = AddDelegate()
        self.returning_delegate = ReturningDelegate()
        self.on_conflict_delegate = OnConflictDelegate()
        self.add(*instances)

    ###########################################################################
    # Clauses

    def add(self: Self, *instances: Table) -> Self:
        self.add_delegate.add(*instances, table_class=self.table)
        return self

    def returning(self: Self, *columns: Column) -> Self:
        self.returning_delegate.returning(columns)
        return self

    def on_conflict(
        self: Self,
        target: t.Optional[t.Union[str, Column, t.Tuple[Column, ...]]] = None,
        action: t.Union[
            OnConflictAction, t.Literal["DO NOTHING", "DO UPDATE"]
        ] = OnConflictAction.do_nothing,
        values: t.Optional[
            t.Sequence[t.Union[Column, t.Tuple[Column, t.Any]]]
        ] = None,
        where: t.Optional[Combinable] = None,
    ) -> Self:
        if (
            self.engine_type == "sqlite"
            and self.table._meta.db.get_version_sync() < 3.24
        ):
            raise NotImplementedError(
                "SQLite versions lower than 3.24 don't support ON CONFLICT"
            )

        if (
            self.engine_type in ("postgres", "cockroach")
            and len(self.on_conflict_delegate._on_conflict.on_conflict_items)
            == 1
        ):
            raise NotImplementedError(
                "Postgres and Cockroach only support a single ON CONFLICT "
                "clause."
            )

        self.on_conflict_delegate.on_conflict(
            target=target,
            action=action,
            values=values,
            where=where,
        )
        return self

    ###########################################################################

    def _raw_response_callback(self, results: t.List):
        """
        Assign the ids of the created rows to the model instances.
        """
        for index, row in enumerate(results):
            table_instance: Table = self.add_delegate._add[index]
            setattr(
                table_instance,
                self.table._meta.primary_key._meta.name,
                row.get(
                    self.table._meta.primary_key._meta.db_column_name, None
                ),
            )
            table_instance._exists_in_db = True

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        base = f"INSERT INTO {self.table._meta.get_formatted_tablename()}"
        columns = ",".join(
            f'"{i._meta.db_column_name}"' for i in self.table._meta.columns
        )
        values = ",".join("{}" for _ in self.add_delegate._add)
        query = f"{base} ({columns}) VALUES {values}"
        querystring = QueryString(
            query,
            *[i.querystring for i in self.add_delegate._add],
            query_type="insert",
            table=self.table,
        )

        engine_type = self.engine_type

        on_conflict = self.on_conflict_delegate._on_conflict
        if on_conflict.on_conflict_items:
            querystring = QueryString(
                "{}{}",
                querystring,
                on_conflict.querystring,
                query_type="insert",
                table=self.table,
            )

        if engine_type in ("postgres", "cockroach") or (
            engine_type == "sqlite"
            and self.table._meta.db.get_version_sync() >= 3.35
        ):
            returning = self.returning_delegate._returning
            if returning:
                return [
                    QueryString(
                        "{}{}",
                        querystring,
                        returning.querystring,
                        query_type="insert",
                        table=self.table,
                    )
                ]

        return [querystring]


Self = t.TypeVar("Self", bound=Insert)



"""
A User model, used for authentication.
"""

from __future__ import annotations

import datetime
import hashlib
import logging
import secrets
import typing as t

from piccolo.columns import Boolean, Secret, Timestamp, Varchar
from piccolo.columns.column_types import Serial
from piccolo.columns.readable import Readable
from piccolo.table import Table
from piccolo.utils.sync import run_sync

logger = logging.getLogger(__name__)


class BaseUser(Table, tablename="piccolo_user"):
    """
    Provides a basic user, with authentication support.
    """

    id: Serial
    username = Varchar(length=100, unique=True)
    password = Secret(length=255)
    first_name = Varchar(null=True)
    last_name = Varchar(null=True)
    email = Varchar(length=255, unique=True)
    active = Boolean(default=False)
    admin = Boolean(
        default=False, help_text="An admin can log into the Piccolo admin GUI."
    )
    superuser = Boolean(
        default=False,
        help_text=(
            "If True, this user can manage other users's passwords in the "
            "Piccolo admin GUI."
        ),
    )
    last_login = Timestamp(
        null=True,
        default=None,
        required=False,
        help_text="When this user last logged in.",
    )

    _min_password_length = 6
    _max_password_length = 128
    # The number of hash iterations recommended by OWASP:
    # https://cheatsheetseries.owasp.org/cheatsheets/Password_Storage_Cheat_Sheet.html#pbkdf2
    _pbkdf2_iteration_count = 600_000

    def __init__(self, **kwargs):
        # Generating passwords upfront is expensive, so might need reworking.
        password = kwargs.get("password", None)
        if password:
            if not password.startswith("pbkdf2_sha256"):
                kwargs["password"] = self.__class__.hash_password(password)
        super().__init__(**kwargs)

    @classmethod
    def get_salt(cls):
        return secrets.token_hex(16)

    @classmethod
    def get_readable(cls) -> Readable:
        """
        Used to get a readable string, representing a table row.
        """
        return Readable(template="%s", columns=[cls.username])

    ###########################################################################

    @classmethod
    def _validate_password(cls, password: str):
        """
        Validate the raw password. Used by :meth:`update_password` and
        :meth:`create_user`.

        :param password:
            The raw password e.g. ``'hello123'``.
        :raises ValueError:
            If the password fails any of the criteria.

        """
        if not password:
            raise ValueError("A password must be provided.")

        if len(password) < cls._min_password_length:
            raise ValueError(
                f"The password is too short. (min {cls._min_password_length})"
            )

        if len(password) > cls._max_password_length:
            raise ValueError(
                f"The password is too long. (max {cls._max_password_length})"
            )

        if password.startswith("pbkdf2_sha256"):
            logger.warning(
                "Tried to create a user with an already hashed password."
            )
            raise ValueError("Do not pass a hashed password.")

    ###########################################################################

    @classmethod
    def update_password_sync(cls, user: t.Union[str, int], password: str):
        """
        A sync equivalent of :meth:`update_password`.
        """
        return run_sync(cls.update_password(user, password))

    @classmethod
    async def update_password(cls, user: t.Union[str, int], password: str):
        """
        The password is the raw password string e.g. ``'password123'``.
        The user can be a user ID, or a username.
        """
        if isinstance(user, str):
            clause = cls.username == user
        elif isinstance(user, int):
            clause = cls.id == user
        else:
            raise ValueError(
                "The `user` arg must be a user id, or a username."
            )

        cls._validate_password(password=password)

        password = cls.hash_password(password)
        await cls.update({cls.password: password}).where(clause).run()

    ###########################################################################

    @classmethod
    def hash_password(
        cls, password: str, salt: str = "", iterations: t.Optional[int] = None
    ) -> str:
        """
        Hashes the password, ready for storage, and for comparing during
        login.

        :raises ValueError:
            If an excessively long password is provided.

        """
        if len(password) > cls._max_password_length:
            logger.warning("Excessively long password provided.")
            raise ValueError("The password is too long.")

        if not salt:
            salt = cls.get_salt()

        if iterations is None:
            iterations = cls._pbkdf2_iteration_count

        hashed = hashlib.pbkdf2_hmac(
            "sha256",
            bytes(password, encoding="utf-8"),
            bytes(salt, encoding="utf-8"),
            iterations,
        ).hex()
        return f"pbkdf2_sha256${iterations}${salt}${hashed}"

    def __setattr__(self, name: str, value: t.Any):
        """
        Make sure that if the password is set, it's stored in a hashed form.
        """
        if name == "password" and not value.startswith("pbkdf2_sha256"):
            value = self.__class__.hash_password(value)

        super().__setattr__(name, value)

    @classmethod
    def split_stored_password(cls, password: str) -> t.List[str]:
        elements = password.split("$")
        if len(elements) != 4:
            raise ValueError("Unable to split hashed password")
        return elements

    ###########################################################################

    @classmethod
    def login_sync(cls, username: str, password: str) -> t.Optional[int]:
        """
        A sync equivalent of :meth:`login`.
        """
        return run_sync(cls.login(username, password))

    @classmethod
    async def login(cls, username: str, password: str) -> t.Optional[int]:
        """
        Make sure the user exists and the password is valid. If so, the
        ``last_login`` value is updated in the database.

        :returns:
            The id of the user if a match is found, otherwise ``None``.

        """
        if len(username) > cls.username.length:
            logger.warning("Excessively long username provided.")
            return None

        if len(password) > cls._max_password_length:
            logger.warning("Excessively long password provided.")
            return None

        response = (
            await cls.select(cls._meta.primary_key, cls.password)
            .where(cls.username == username)
            .first()
            .run()
        )
        if not response:
            # No match found. We still call hash_password
            # here to mitigate the ability to enumerate
            # users via response timings
            cls.hash_password(password)
            return None

        stored_password = response["password"]

        algorithm, iterations_, salt, hashed = cls.split_stored_password(
            stored_password
        )
        iterations = int(iterations_)

        if cls.hash_password(password, salt, iterations) == stored_password:
            # If the password was hashed in an earlier Piccolo version, update
            # it so it's hashed with the currently recommended number of
            # iterations:
            if iterations != cls._pbkdf2_iteration_count:
                await cls.update_password(username, password)

            await cls.update({cls.last_login: datetime.datetime.now()}).where(
                cls.username == username
            )
            return response["id"]
        else:
            return None

    ###########################################################################

    @classmethod
    def create_user_sync(
        cls, username: str, password: str, **extra_params
    ) -> BaseUser:
        """
        A sync equivalent of :meth:`create_user`.
        """
        return run_sync(
            cls.create_user(
                username=username, password=password, **extra_params
            )
        )

    @classmethod
    async def create_user(
        cls, username: str, password: str, **extra_params
    ) -> BaseUser:
        """
        Creates a new user, and saves it in the database. It is recommended to
        use this rather than instantiating and saving ``BaseUser`` directly, as
        we add extra validation.

        :raises ValueError:
            If the username or password is invalid.
        :returns:
            The created ``BaseUser`` instance.

        """
        if not username:
            raise ValueError("A username must be provided.")

        cls._validate_password(password=password)

        user = cls(username=username, password=password, **extra_params)
        await user.save()
        return user



from __future__ import annotations

import typing as t

from piccolo.custom_types import Combinable, TableInstance
from piccolo.query.base import Query
from piccolo.query.mixins import (
    ReturningDelegate,
    ValuesDelegate,
    WhereDelegate,
)
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.columns import Column


class UpdateError(Exception):
    pass


class Update(Query[TableInstance, t.List[t.Any]]):
    __slots__ = (
        "force",
        "returning_delegate",
        "values_delegate",
        "where_delegate",
    )

    def __init__(
        self, table: t.Type[TableInstance], force: bool = False, **kwargs
    ):
        super().__init__(table, **kwargs)
        self.force = force
        self.returning_delegate = ReturningDelegate()
        self.values_delegate = ValuesDelegate(table=table)
        self.where_delegate = WhereDelegate()

    ###########################################################################
    # Clauses

    def values(
        self,
        values: t.Optional[t.Dict[t.Union[Column, str], t.Any]] = None,
        **kwargs,
    ) -> Update:
        if values is None:
            values = {}
        values = dict(values, **kwargs)
        self.values_delegate.values(values)
        return self

    def where(self, *where: t.Union[Combinable, QueryString]) -> Update:
        self.where_delegate.where(*where)
        return self

    def returning(self, *columns: Column) -> Update:
        self.returning_delegate.returning(columns)
        return self

    ###########################################################################

    def _validate(self):
        """
        Called at the start of :meth:`piccolo.query.base.Query.run` to make
        sure the user has configured the query correctly before running it.
        """
        if len(self.values_delegate._values) == 0:
            raise ValueError("No values were specified to update.")

        for column, _ in self.values_delegate._values.items():
            if len(column._meta.call_chain) > 0:
                raise ValueError(
                    "Related values can't be updated via an update."
                )

        if (not self.where_delegate._where) and (not self.force):
            classname = self.table.__name__
            raise UpdateError(
                "Do you really want to update all rows in "
                f"{classname}? If so, use pass `force=True` into "
                f"`{classname}.update`. Otherwise, add a where clause."
            )

    ###########################################################################

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        columns_str = ", ".join(
            f'"{col._meta.db_column_name}" = {{}}'
            for col, _ in self.values_delegate._values.items()
        )

        query = f"UPDATE {self.table._meta.get_formatted_tablename()} SET {columns_str}"  # noqa: E501

        querystring = QueryString(
            query, *self.values_delegate.get_sql_values()
        )

        if self.where_delegate._where:
            # The JOIN syntax isn't allowed in SQL UPDATE queries, so we need
            # to write the WHERE clause differently, using a sub select.

            querystring = QueryString(
                "{} WHERE {}",
                querystring,
                self.where_delegate._where.querystring_for_update,
            )

        if self.returning_delegate._returning:
            querystring = QueryString(
                "{}{}",
                querystring,
                self.returning_delegate._returning.querystring,
            )

        return [querystring]



from __future__ import annotations

import typing as t

from piccolo.engine.base import BaseBatch
from piccolo.query.base import Query
from piccolo.querystring import QueryString

if t.TYPE_CHECKING:  # pragma: no cover
    from piccolo.table import Table


class Raw(Query):
    __slots__ = ("querystring",)

    def __init__(
        self,
        table: t.Type[Table],
        querystring: QueryString = QueryString(""),
        **kwargs,
    ):
        super().__init__(table, **kwargs)
        self.querystring = querystring

    async def batch(
        self,
        batch_size: t.Optional[int] = None,
        node: t.Optional[str] = None,
        **kwargs,
    ) -> BaseBatch:
        if batch_size:
            kwargs.update(batch_size=batch_size)
        if node:
            kwargs.update(node=node)
        return await self.table._meta.db.batch(self, **kwargs)

    @property
    def default_querystrings(self) -> t.Sequence[QueryString]:
        return [self.querystring]



from __future__ import annotations

import contextvars
import datetime
import enum
import os
import sqlite3
import typing as t
import uuid
from dataclasses import dataclass
from decimal import Decimal
from functools import partial, wraps

from typing_extensions import Self

from piccolo.engine.base import (
    BaseAtomic,
    BaseBatch,
    BaseTransaction,
    Engine,
    validate_savepoint_name,
)
from piccolo.engine.exceptions import TransactionError
from piccolo.query.base import DDL, Query
from piccolo.querystring import QueryString
from piccolo.utils.encoding import dump_json, load_json
from piccolo.utils.lazy_loader import LazyLoader
from piccolo.utils.sync import run_sync

aiosqlite = LazyLoader("aiosqlite", globals(), "aiosqlite")


if t.TYPE_CHECKING:  # pragma: no cover
    from aiosqlite import Connection, Cursor  # type: ignore

    from piccolo.table import Table

###############################################################################

# We need to register some adapters so sqlite returns types which are more
# consistent with the Postgres engine.


# In


def convert_numeric_in(value: Decimal) -> float:
    """
    Convert any Decimal values into floats.
    """
    return float(value)


def convert_uuid_in(value: uuid.UUID) -> str:
    """
    Converts the UUID value being passed into sqlite.
    """
    return str(value)


def convert_time_in(value: datetime.time) -> str:
    """
    Converts the time value being passed into sqlite.
    """
    return value.isoformat()


def convert_date_in(value: datetime.date) -> str:
    """
    Converts the date value being passed into sqlite.
    """
    return value.isoformat()


def convert_datetime_in(value: datetime.datetime) -> str:
    """
    Converts the datetime into a string. If it's timezone aware, we want to
    convert it to UTC first. This is to replicate Postgres, which stores
    timezone aware datetimes in UTC.
    """
    if value.tzinfo is not None:
        value = value.astimezone(datetime.timezone.utc)
    return str(value)


def convert_timedelta_in(value: datetime.timedelta) -> float:
    """
    Converts the timedelta value being passed into sqlite.
    """
    return value.total_seconds()


def convert_array_in(value: list) -> str:
    """
    Converts a list value into a string (it handles nested lists, and type like
    dateime/ time / date which aren't usually JSON serialisable.).

    """

    def serialise(data: list):
        output = []

        for item in data:
            if isinstance(item, list):
                output.append(serialise(item))
            elif isinstance(
                item, (datetime.datetime, datetime.time, datetime.date)
            ):
                if adapter := ADAPTERS.get(type(item)):
                    output.append(adapter(item))
                else:
                    raise ValueError("The adapter wasn't found.")
            elif item is None or isinstance(item, (str, int, float, list)):
                # We can safely JSON serialise these.
                output.append(item)
            else:
                raise ValueError("We can't currently serialise this value.")

        return output

    return dump_json(serialise(value))


###############################################################################

# Register adapters

ADAPTERS: t.Dict[t.Type, t.Callable[[t.Any], t.Any]] = {
    Decimal: convert_numeric_in,
    uuid.UUID: convert_uuid_in,
    datetime.time: convert_time_in,
    datetime.date: convert_date_in,
    datetime.datetime: convert_datetime_in,
    datetime.timedelta: convert_timedelta_in,
    list: convert_array_in,
}

for value_type, adapter in ADAPTERS.items():
    sqlite3.register_adapter(value_type, adapter)

###############################################################################

# Out


def decode_to_string(converter: t.Callable[[str], t.Any]):
    """
    This means we can use our converters with string and bytes. They are
    passed bytes when used directly via SQLite, and are passed strings when
    used by the array converters.
    """

    @wraps(converter)
    def wrapper(value: t.Union[str, bytes]) -> t.Any:
        if isinstance(value, bytes):
            return converter(value.decode("utf8"))
        elif isinstance(value, str):
            return converter(value)
        else:
            raise ValueError("Unsupported type")

    return wrapper


@decode_to_string
def convert_numeric_out(value: str) -> Decimal:
    """
    Convert float values into Decimals.
    """
    return Decimal(value)


@decode_to_string
def convert_int_out(value: str) -> int:
    """
    Make sure Integer values are actually of type int.
    """
    return int(float(value))


@decode_to_string
def convert_uuid_out(value: str) -> uuid.UUID:
    """
    If the value is a uuid, convert it to a UUID instance.
    """
    return uuid.UUID(value)


@decode_to_string
def convert_date_out(value: str) -> datetime.date:
    return datetime.date.fromisoformat(value)


@decode_to_string
def convert_time_out(value: str) -> datetime.time:
    """
    If the value is a time, convert it to a UUID instance.
    """
    return datetime.time.fromisoformat(value)


@decode_to_string
def convert_seconds_out(value: str) -> datetime.timedelta:
    """
    If the value is from a seconds column, convert it to a timedelta instance.
    """
    return datetime.timedelta(seconds=float(value))


@decode_to_string
def convert_boolean_out(value: str) -> bool:
    """
    If the value is from a boolean column, convert it to a bool value.
    """
    return value == "1"


@decode_to_string
def convert_timestamp_out(value: str) -> datetime.datetime:
    """
    If the value is from a timestamp column, convert it to a datetime value.
    """
    return datetime.datetime.fromisoformat(value)


@decode_to_string
def convert_timestamptz_out(value: str) -> datetime.datetime:
    """
    If the value is from a timestamptz column, convert it to a datetime value,
    with a timezone of UTC.
    """
    return datetime.datetime.fromisoformat(value).replace(
        tzinfo=datetime.timezone.utc
    )


@decode_to_string
def convert_array_out(value: str) -> t.List:
    """
    If the value if from an array column, deserialise the string back into a
    list.
    """
    return load_json(value)


def convert_complex_array_out(value: bytes, converter: t.Callable):
    """
    This is used to handle arrays of things like timestamps, which we can't
    just load from JSON without doing additional work to convert the elements
    back into Python objects.
    """
    parsed = load_json(value.decode("utf8"))

    def convert_list(list_value: t.List):
        output = []

        for value in list_value:
            if isinstance(value, list):
                # For nested arrays
                output.append(convert_list(value))
            elif isinstance(value, str):
                output.append(converter(value))
            else:
                output.append(value)

        return output

    if isinstance(parsed, list):
        return convert_list(parsed)
    else:
        return parsed


@decode_to_string
def convert_M2M_out(value: str) -> t.List:
    return value.split(",")


###############################################################################
# Register the basic converters

CONVERTERS = {
    "NUMERIC": convert_numeric_out,
    "INTEGER": convert_int_out,
    "UUID": convert_uuid_out,
    "DATE": convert_date_out,
    "TIME": convert_time_out,
    "SECONDS": convert_seconds_out,
    "BOOLEAN": convert_boolean_out,
    "TIMESTAMP": convert_timestamp_out,
    "TIMESTAMPTZ": convert_timestamptz_out,
    "M2M": convert_M2M_out,
}

for column_name, converter in CONVERTERS.items():
    sqlite3.register_converter(column_name, converter)

###############################################################################
# Register the array converters

# The ARRAY column type handles values which can be easily serialised to and
# from JSON.
sqlite3.register_converter("ARRAY", convert_array_out)

# We have special column types for arrays of timestamps etc, as simply loading
# the JSON isn't sufficient.
for column_name in ("TIMESTAMP", "TIMESTAMPTZ", "DATE", "TIME"):
    sqlite3.register_converter(
        f"ARRAY_{column_name}",
        partial(
            convert_complex_array_out,
            converter=CONVERTERS[column_name],
        ),
    )

###############################################################################


@dataclass
class AsyncBatch(BaseBatch):
    connection: Connection
    query: Query
    batch_size: int

    # Set internally
    _cursor: t.Optional[Cursor] = None

    @property
    def cursor(self) -> Cursor:
        if not self._cursor:
            raise ValueError("_cursor not set")
        return self._cursor

    async def next(self) -> t.List[t.Dict]:
        data = await self.cursor.fetchmany(self.batch_size)
        return await self.query._process_results(data)

    def __aiter__(self: Self) -> Self:
        return self

    async def __anext__(self) -> t.List[t.Dict]:
        response = await self.next()
        if response == []:
            raise StopAsyncIteration()
        return response

    async def __aenter__(self: Self) -> Self:
        querystring = self.query.querystrings[0]
        template, template_args = querystring.compile_string()

        self._cursor = await self.connection.execute(template, *template_args)
        return self

    async def __aexit__(self, exception_type, exception, traceback):
        await self.cursor.close()
        await self.connection.close()
        return exception is not None


###############################################################################


class TransactionType(enum.Enum):
    """
    See the `SQLite <https://www.sqlite.org/lang_transaction.html>`_ docs for
    more info.
    """

    deferred = "DEFERRED"
    immediate = "IMMEDIATE"
    exclusive = "EXCLUSIVE"


class Atomic(BaseAtomic):
    """
    Usage:

    transaction = engine.atomic()
    transaction.add(Foo.create_table())

    # Either:
    transaction.run_sync()
    await transaction.run()
    """

    __slots__ = ("engine", "queries", "transaction_type")

    def __init__(
        self,
        engine: SQLiteEngine,
        transaction_type: TransactionType = TransactionType.deferred,
    ):
        self.engine = engine
        self.transaction_type = transaction_type
        self.queries: t.List[t.Union[Query, DDL]] = []

    def add(self, *query: t.Union[Query, DDL]):
        self.queries += list(query)

    async def run(self):
        from piccolo.query.methods.objects import Create, GetOrCreate

        try:
            async with self.engine.transaction(
                transaction_type=self.transaction_type
            ):
                for query in self.queries:
                    if isinstance(query, (Query, DDL, Create, GetOrCreate)):
                        await query.run()
                    else:
                        raise ValueError("Unrecognised query")
            self.queries = []
        except Exception as exception:
            self.queries = []
            raise exception from exception

    def run_sync(self):
        return run_sync(self.run())

    def __await__(self):
        return self.run().__await__()


###############################################################################


class Savepoint:
    def __init__(self, name: str, transaction: SQLiteTransaction):
        self.name = name
        self.transaction = transaction

    async def rollback_to(self):
        validate_savepoint_name(self.name)
        await self.transaction.connection.execute(
            f"ROLLBACK TO SAVEPOINT {self.name}"
        )

    async def release(self):
        validate_savepoint_name(self.name)
        await self.transaction.connection.execute(
            f"RELEASE SAVEPOINT {self.name}"
        )


class SQLiteTransaction(BaseTransaction):
    """
    Used for wrapping queries in a transaction, using a context manager.
    Currently it's async only.

    Usage::

        async with engine.transaction():
            # Run some queries:
            await Band.select().run()

    """

    __slots__ = (
        "engine",
        "context",
        "connection",
        "transaction_type",
        "allow_nested",
        "_savepoint_id",
        "_parent",
        "_committed",
        "_rolled_back",
    )

    def __init__(
        self,
        engine: SQLiteEngine,
        transaction_type: TransactionType = TransactionType.deferred,
        allow_nested: bool = True,
    ):
        """
        :param transaction_type:
            If your transaction just contains ``SELECT`` queries, then use
            ``TransactionType.deferred``. This will give you the best
            performance. When performing ``INSERT``, ``UPDATE``, ``DELETE``
            queries, we recommend using ``TransactionType.immediate`` to
            avoid database locks.
        """
        self.engine = engine
        self.transaction_type = transaction_type
        current_transaction = self.engine.current_transaction.get()

        self._savepoint_id = 0
        self._parent = None
        self._committed = False
        self._rolled_back = False

        if current_transaction:
            if allow_nested:
                self._parent = current_transaction
            else:
                raise TransactionError(
                    "A transaction is already active - nested transactions "
                    "aren't allowed."
                )

    async def __aenter__(self) -> SQLiteTransaction:
        if self._parent is not None:
            return self._parent

        self.connection = await self.get_connection()
        await self.begin()
        self.context = self.engine.current_transaction.set(self)
        return self

    async def get_connection(self):
        return await self.engine.get_connection()

    async def begin(self):
        await self.connection.execute(f"BEGIN {self.transaction_type.value}")

    async def commit(self):
        await self.connection.execute("COMMIT")
        self._committed = True

    async def rollback(self):
        await self.connection.execute("ROLLBACK")
        self._rolled_back = True

    async def rollback_to(self, savepoint_name: str):
        """
        Used to rollback to a savepoint just using the name.
        """
        await Savepoint(name=savepoint_name, transaction=self).rollback_to()

    ###########################################################################

    def get_savepoint_id(self) -> int:
        self._savepoint_id += 1
        return self._savepoint_id

    async def savepoint(self, name: t.Optional[str] = None) -> Savepoint:
        name = name or f"savepoint_{self.get_savepoint_id()}"
        validate_savepoint_name(name)
        await self.connection.execute(f"SAVEPOINT {name}")
        return Savepoint(name=name, transaction=self)

    ###########################################################################

    async def __aexit__(self, exception_type, exception, traceback) -> bool:
        if self._parent:
            return exception is None

        if exception:
            # The user may have manually rolled it back.
            if not self._rolled_back:
                await self.rollback()
        else:
            # The user may have manually committed it.
            if not self._committed and not self._rolled_back:
                await self.commit()

        await self.connection.close()
        self.engine.current_transaction.reset(self.context)

        return exception is None


###############################################################################


def dict_factory(cursor, row) -> t.Dict:
    return {col[0]: row[idx] for idx, col in enumerate(cursor.description)}


class SQLiteEngine(Engine[SQLiteTransaction]):
    __slots__ = ("connection_kwargs",)

    def __init__(
        self,
        path: str = "piccolo.sqlite",
        log_queries: bool = False,
        log_responses: bool = False,
        **connection_kwargs,
    ) -> None:
        """
        :param path:
            A relative or absolute path to the the SQLite database file (it
            will be created if it doesn't already exist).
        :param log_queries:
            If ``True``, all SQL and DDL statements are printed out before
            being run. Useful for debugging.
        :param log_responses:
            If ``True``, the raw response from each query is printed out.
            Useful for debugging.
        :param connection_kwargs:
            These are passed directly to the database adapter. We recommend
            setting ``timeout`` if you expect your application to process a
            large number of concurrent writes, to prevent queries timing out.
            See Python's `sqlite3 docs <https://docs.python.org/3/library/sqlite3.html#sqlite3.connect>`_
            for more info.

        """  # noqa: E501
        default_connection_kwargs = {
            "database": path,
            "detect_types": sqlite3.PARSE_DECLTYPES | sqlite3.PARSE_COLNAMES,
            "isolation_level": None,
        }

        self.log_queries = log_queries
        self.log_responses = log_responses
        self.connection_kwargs = {
            **default_connection_kwargs,
            **connection_kwargs,
        }

        self.current_transaction = contextvars.ContextVar(
            f"sqlite_current_transaction_{path}", default=None
        )

        super().__init__(
            engine_type="sqlite",
            min_version_number=3.25,
            log_queries=log_queries,
            log_responses=log_responses,
        )

    @property
    def path(self):
        return self.connection_kwargs["database"]

    @path.setter
    def path(self, value: str):
        self.connection_kwargs["database"] = value

    async def get_version(self) -> float:
        return self.get_version_sync()

    def get_version_sync(self) -> float:
        major, minor, _ = sqlite3.sqlite_version_info
        return float(f"{major}.{minor}")

    async def prep_database(self):
        pass

    ###########################################################################

    def remove_db_file(self):
        """
        Use with caution - removes the SQLite file. Useful for testing
        purposes.
        """
        if os.path.exists(self.path):
            os.unlink(self.path)

    def create_db_file(self):
        """
        Create the database file. Useful for testing purposes.
        """
        if os.path.exists(self.path):
            raise Exception(f"Database at {self.path} already exists")
        with open(self.path, "w"):
            pass

    ###########################################################################

    async def batch(
        self, query: Query, batch_size: int = 100, node: t.Optional[str] = None
    ) -> AsyncBatch:
        """
        :param query:
            The database query to run.
        :param batch_size:
            How many rows to fetch on each iteration.
        :param node:
            This is ignored currently, as SQLite runs off a single node. The
            value is here so the API is consistent with Postgres.
        """
        connection = await self.get_connection()
        return AsyncBatch(
            connection=connection, query=query, batch_size=batch_size
        )

    ###########################################################################

    async def get_connection(self) -> Connection:
        connection = await aiosqlite.connect(**self.connection_kwargs)
        connection.row_factory = dict_factory  # type: ignore
        await connection.execute("PRAGMA foreign_keys = 1")
        return connection

    ###########################################################################

    async def _get_inserted_pk(self, cursor, table: t.Type[Table]) -> t.Any:
        """
        If the `pk` column is a non-integer then `ROWID` and `pk` will return
        different types. Need to query by `lastrowid` to get `pk`s in SQLite
        prior to 3.35.0.
        """
        await cursor.execute(
            f"SELECT {table._meta.primary_key._meta.db_column_name} FROM "
            f"{table._meta.tablename} WHERE ROWID = {cursor.lastrowid}"
        )
        response = await cursor.fetchone()
        return response[table._meta.primary_key._meta.db_column_name]

    async def _run_in_new_connection(
        self,
        query: str,
        args: t.Optional[t.List[t.Any]] = None,
        query_type: str = "generic",
        table: t.Optional[t.Type[Table]] = None,
    ):
        if args is None:
            args = []
        async with aiosqlite.connect(**self.connection_kwargs) as connection:
            await connection.execute("PRAGMA foreign_keys = 1")

            connection.row_factory = dict_factory  # type: ignore
            async with connection.execute(query, args) as cursor:
                await connection.commit()

                if query_type == "insert" and self.get_version_sync() < 3.35:
                    # We can't use the RETURNING clause on older versions
                    # of SQLite.
                    assert table is not None
                    pk = await self._get_inserted_pk(cursor, table)
                    return [{table._meta.primary_key._meta.db_column_name: pk}]
                else:
                    return await cursor.fetchall()

    async def _run_in_existing_connection(
        self,
        connection,
        query: str,
        args: t.Optional[t.List[t.Any]] = None,
        query_type: str = "generic",
        table: t.Optional[t.Type[Table]] = None,
    ):
        """
        This is used when a transaction is currently active.
        """
        if args is None:
            args = []
        await connection.execute("PRAGMA foreign_keys = 1")

        connection.row_factory = dict_factory
        async with connection.execute(query, args) as cursor:
            response = await cursor.fetchall()

            if query_type == "insert" and self.get_version_sync() < 3.35:
                # We can't use the RETURNING clause on older versions
                # of SQLite.
                assert table is not None
                pk = await self._get_inserted_pk(cursor, table)
                return [{table._meta.primary_key._meta.db_column_name: pk}]
            else:
                return response

    async def run_querystring(
        self, querystring: QueryString, in_pool: bool = False
    ):
        """
        Connection pools aren't currently supported - the argument is there
        for consistency with other engines.
        """
        query_id = self.get_query_id()

        if self.log_queries:
            self.print_query(query_id=query_id, query=querystring.__str__())

        query, query_args = querystring.compile_string(
            engine_type=self.engine_type
        )

        # If running inside a transaction:
        current_transaction = self.current_transaction.get()
        if current_transaction:
            response = await self._run_in_existing_connection(
                connection=current_transaction.connection,
                query=query,
                args=query_args,
                query_type=querystring.query_type,
                table=querystring.table,
            )
        else:
            response = await self._run_in_new_connection(
                query=query,
                args=query_args,
                query_type=querystring.query_type,
                table=querystring.table,
            )

        if self.log_responses:
            self.print_response(query_id=query_id, response=response)

        return response

    async def run_ddl(self, ddl: str, in_pool: bool = False):
        """
        Connection pools aren't currently supported - the argument is there
        for consistency with other engines.
        """
        query_id = self.get_query_id()

        if self.log_queries:
            self.print_query(query_id=query_id, query=ddl)

        # If running inside a transaction:
        current_transaction = self.current_transaction.get()
        if current_transaction:
            response = await self._run_in_existing_connection(
                connection=current_transaction.connection,
                query=ddl,
            )
        else:
            response = await self._run_in_new_connection(
                query=ddl,
            )

        if self.log_responses:
            self.print_response(query_id=query_id, response=response)

        return response

    def atomic(
        self, transaction_type: TransactionType = TransactionType.deferred
    ) -> Atomic:
        return Atomic(engine=self, transaction_type=transaction_type)

    def transaction(
        self,
        transaction_type: TransactionType = TransactionType.deferred,
        allow_nested: bool = True,
    ) -> SQLiteTransaction:
        """
        Create a new database transaction. See :class:`Transaction`.
        """
        return SQLiteTransaction(
            engine=self,
            transaction_type=transaction_type,
            allow_nested=allow_nested,
        )



import typing as t

from piccolo.testing.test_case import AsyncTableTest
from tests.example_apps.music.tables import Band, Concert, Manager, Venue


class TestGetRelated(AsyncTableTest):
    tables = [Manager, Band, Concert, Venue]

    async def asyncSetUp(self):
        await super().asyncSetUp()

        # Setup two pairs of manager/band, so we can make sure the correct
        # objects are returned.

        self.manager = Manager(name="Guido")
        await self.manager.save()

        self.band = Band(
            name="Pythonistas", manager=self.manager.id, popularity=100
        )
        await self.band.save()

        self.manager_2 = Manager(name="Graydon")
        await self.manager_2.save()

        self.band_2 = Band(
            name="Rustaceans", manager=self.manager_2.id, popularity=100
        )
        await self.band_2.save()

    async def test_foreign_key(self) -> None:
        """
        Make sure you can get a related object from another object instance.
        """
        manager = await self.band.get_related(Band.manager)
        assert manager is not None
        self.assertTrue(manager.id == self.manager.id)

        manager_2 = await self.band_2.get_related(Band.manager)
        assert manager_2 is not None
        self.assertTrue(manager_2.id == self.manager_2.id)

    async def test_non_foreign_key(self):
        """
        Make sure that non-ForeignKey raise an exception.
        """
        with self.assertRaises(ValueError):
            self.band.get_related(Band.name)  # type: ignore

    async def test_string(self):
        """
        Make sure it also works using a string representation of a foreign key.
        """
        manager = t.cast(Manager, await self.band.get_related("manager"))
        self.assertTrue(manager.id == self.manager.id)

    async def test_invalid_string(self):
        """
        Make sure an exception is raised if the foreign key string is invalid.
        """
        with self.assertRaises(ValueError):
            self.band.get_related("abc123")

    async def test_multiple_levels(self):
        """
        Make sure ``get_related`` works multiple levels deep.
        """
        concert = Concert(band_1=self.band, band_2=self.band_2)
        await concert.save()

        manager = await concert.get_related(Concert.band_1._.manager)
        assert manager is not None
        self.assertTrue(manager.id == self.manager.id)

        manager_2 = await concert.get_related(Concert.band_2._.manager)
        assert manager_2 is not None
        self.assertTrue(manager_2.id == self.manager_2.id)

    async def test_no_match(self):
        """
        If not related object exists, make sure ``None`` is returned.
        """
        concert = Concert(band_1=self.band, band_2=None)
        await concert.save()

        manager_2 = await concert.get_related(Concert.band_2._.manager)
        assert manager_2 is None

    async def test_not_in_db(self):
        """
        If the object we're calling ``get_related`` on doesn't exist in the
        database, then make sure an error is raised.
        """
        concert = Concert(band_1=self.band, band_2=self.band_2)

        with self.assertRaises(ValueError):
            await concert.get_related(Concert.band_1._.manager)



from __future__ import annotations

import typing as t
from unittest import IsolatedAsyncioTestCase, TestCase

from piccolo.engine import Engine, engine_finder
from piccolo.table import (
    Table,
    create_db_tables,
    create_db_tables_sync,
    drop_db_tables,
    drop_db_tables_sync,
)


class TableTest(TestCase):
    """
    Identical to :class:`AsyncTableTest <piccolo.testing.test_case.AsyncTableTest>`,
    except it only work for sync tests. Only use this if you can't make your
    tests async (perhaps you're on Python 3.7 where ``IsolatedAsyncioTestCase``
    isn't available).

    For example::

        class TestBand(TableTest):
            tables = [Band]

            def test_band(self):
                ...

    """  # noqa: E501

    tables: t.List[t.Type[Table]]

    def setUp(self) -> None:
        create_db_tables_sync(*self.tables)

    def tearDown(self) -> None:
        drop_db_tables_sync(*self.tables)


class AsyncTableTest(IsolatedAsyncioTestCase):
    """
    Used for tests where we need to create Piccolo tables - they will
    automatically be created and dropped.

    For example::

        class TestBand(AsyncTableTest):
            tables = [Band]

            async def test_band(self):
                ...

    """

    tables: t.List[t.Type[Table]]

    async def asyncSetUp(self) -> None:
        await create_db_tables(*self.tables)

    async def asyncTearDown(self) -> None:
        await drop_db_tables(*self.tables)


class AsyncTransactionTest(IsolatedAsyncioTestCase):
    """
    Wraps each test in a transaction, which is automatically rolled back when
    the test finishes.

    .. warning::
        Python 3.11 and above only.

    If your test suite just contains ``AsyncTransactionTest`` tests, then you
    can setup your database tables once before your test suite runs. Any
    changes made to your tables by the tests will be rolled back automatically.

    Here's an example::

        from piccolo.testing.test_case import AsyncTransactionTest


        class TestBandEndpoint(AsyncTransactionTest):

            async def test_band_response(self):
                \"\"\"
                Make sure the endpoint returns a 200.
                \"\"\"
                band = Band({Band.name: "Pythonistas"})
                await band.save()

                # Using an API testing client, like httpx:
                response = await client.get(f"/bands/{band.id}/")
                self.assertEqual(response.status_code, 200)

    We add a ``Band`` to the database, but any subsequent tests won't see it,
    as the changes are rolled back automatically.

    """

    # We use `engine_finder` to find the current `Engine`, but you can
    # explicity set it here if you prefer:
    #
    # class MyTest(AsyncTransactionTest):
    #     db = DB
    #
    #     ...
    #
    db: t.Optional[Engine] = None

    async def asyncSetUp(self) -> None:
        db = self.db or engine_finder()
        assert db is not None
        self.transaction = db.transaction()
        # This is only available in Python 3.11 and above:
        await self.enterAsyncContext(cm=self.transaction)  # type: ignore

    async def asyncTearDown(self):
        await super().asyncTearDown()
        await self.transaction.rollback()



from piccolo.testing.test_case import AsyncTableTest
from tests.example_apps.music.tables import Band, Manager


class TestUpdateSelf(AsyncTableTest):

    tables = [Band, Manager]

    async def test_update_self(self):
        band = Band({Band.name: "Pythonistas", Band.popularity: 1000})

        # Make sure we get a ValueError if it's not in the database yet.
        with self.assertRaises(ValueError):
            await band.update_self({Band.popularity: Band.popularity + 1})

        # Save it, so it's in the database
        await band.save()

        # Make sure we can successfully update the object
        await band.update_self({Band.popularity: Band.popularity + 1})

        # Make sure the value was updated on the object
        assert band.popularity == 1001

        # Make sure the value was updated in the database
        await band.refresh()
        assert band.popularity == 1001
